from redis import Redis
import pickle
from io import BytesIO


class BaseList(object):
    """if you want to create a new list just extend this class"""
    conn = None
    list_name = __name__

    def __init__(self, connection=None, list_name=None, job_hash=None):
        assert connection is not None, "the redis connection is required"
        assert job_hash is not None, "the job hash is required"
        self.conn = connection
        self.job_hash = job_hash(self.conn)
        if list_name:
            self.list_name = list_name

    def __iter__(self):
        return self

    def __next__(self):
        if self.len() > 0:
            ret = self.pop()
            return ret
        else:
            raise StopIteration()

    def queue(self, job):
        """queues a job into this list"""
        # TODO serializer job
        job.status = "queue into {0}".format(self.list_name)
        self.job_hash.set(job)
        job = pickle.dumps(job)
        self.conn.lpush(self.list_name, job)

    def pop_to(self, another_list):
        """moves the least job from this list into another list """
        job = self.get(-1)
        if job:
            job.status = 'move form {this_list} into {list_name}'.format(
            this_list=self.list_name,
            list_name=another_list.list_name)
            self.set(-1, job)
        job = self.conn.rpoplpush(self.list_name, another_list.list_name)
        if job is not None:
            job = pickle.loads(job)
            self.job_hash.set(job)
            return job

    def pop(self):
        _job = self.conn.lpop(self.list_name)
        if _job:
            job = pickle.loads(_job)
            self.job_hash.remove(job)
            return job
        else:
            return None

    def pull(self):
        job = pickle.loads(self.conn.rpop(self.list_name))
        self.job_hash.remove(job)
        return job

    def re_queue(self, another_list):
        job = another_list.pull()
        job = pickle.dumps(job)
        self.conn.rpush(self.list_name, job)

    def set(self, index, job):
        self.job_hash.set(job)
        job = pickle.dumps(job)
        self.conn.lset(self.list_name, index, job)

    def get(self, index):
        """returns the job from this list at the index"""
        job = self.conn.lindex(self.list_name, index)
        if job:
            return pickle.loads(job)
        else:
            return None

    def len(self):
        return self.conn.llen(self.list_name)


class HS(object):
    """the hash of holding sign workers worker:working list"""
    hash_name = 'sign_hash'
    conn = None

    def __init__(self, connection=None, hash_name=None):
        assert connection is not None, "the redis connection is required"
        self.conn = connection
        if hash_name:
            self.hash_name = hash_name
    def get_member(self):
        return self.conn.hgetall(self.hash_name)

    def sign_in(self, worker, list):
        self.conn.hset(self.hash_name, worker.id, list.list_name)

    def sign_out(self, worker):
        self.conn.hget(self.hash_name, worker.id)


class HJ(object):
    """the hash of holding job_id:job"""
    hash_name = 'job_hash'
    conn = None

    def __init__(self, connection=None, hash_name=None):
        assert connection is not None, "the redis connection is required"
        self.conn = connection
        if hash_name:
            self.hash_name = hash_name
    def keys(self):
        return self.conn.hgetall(self.hash_name)

    def set(self, job):
        _job = pickle.dumps(job)
        self.conn.hset(self.hash_name, job.id, _job)

    def fetch(self, id):
        return pickle.loads(self.conn.hget(self.hash_name, id))

    def remove(self, job):
        pass


class LQ(BaseList):
    """the list of holding queue jobs"""
    list_name = 'queue_list'


class LW(BaseList):
    """the list of holding working jobs"""
    list_name = 'work_list'


class LD(BaseList):
    """the list of holding done jobs"""
    list_name = 'done_list'

class LF(BaseList):
    """the list of holding failed jobs"""
    list_name = 'failed_list'



import sys
sys.path.append('../../')
from RedisQ.GPUWorker import GPUWorker
from redis import Redis
from RedisQ.Base.Qjob import BaseJob
from RedisQ.Base.Qlist import LQ, LW, LD, HS, HJ, LF
from pyseri.serializer import serializer
from threading import Thread
import json
import csv
import requests
import os


class DICOMPredictionContent(object):
    # those data is used for testing
    api = 'http://192.168.1.150:8000/api/v1/'
    image = 'images/?accession_id=4&limit=200&ordering=id'
    api = 'http://192.168.1.150:8000/api/v1/'
    patient = 'patients/'
    patient_id = '1'
    case = 'cases/'
    case_id = '2351cb63-ea0a-4f40-8fed-5a82761adb7e'
    accession = 'accessions/'
    accession_id = '1'
    cad = 'cads/'


class DICOMJob(BaseJob):

    content = DICOMPredictionContent()


class DICOMConfig(object):

    gpu_load = 0.3
    gpu_memory = 0.3
    docker_path = './test'
    docker_image_tag = 'test_work'
    res_base = 'DICOM/res'
    des_base = 'workout/DICOM/des'

    @property
    def gpu_fraction(self):
        return (self.gpu_memory + self.gpu_load) * 0.5


def payload(job, config):
    print('playload running...')
    # get job
    content = job.content
    # get api information
    base = '{base}/{accession_id}/'.format(
        base=config.res_base,
        accession_id=content.accession_id
    )

    if not os.path.isdir(base):
        print('create path')
        os.mkdir(base)

    print('get resource info')

    response = requests.get(content.api + content.image)

    if response.status_code != 200:
        raise Exception("can't get resource information", response.content)

    for image in response.json().get('results'):
        raw_uri = image.get('raw_file_path', None)
        id = image.get('id', None)
        if raw_uri and id is None:
            raise Exception(
                "can get resource information 'cause raw_uri or id is not in the response")
        # judge if the raw file has been down
        filename = raw_uri.split('/')[-1]
        res = '{base}{id}_{filename}'.format(
            base=base,
            id=id,
            filename=filename
        )
        if not os.path.isfile(res):
            print('download raw file')
            response = requests.get(raw_uri)
            if response.status_code != 200:
                raise Exception(
                    "can't download raw files {},{}".format(raw_uri, id))
        # set file path
        with open(res, 'wb+') as f:
            f.write(response.content)

    print('ready to predict')
    des = '{base}/{accession_id}/'.format(
        base=config.des_base,
        accession_id=content.accession_id
    )
    # write content
    # make destenation path if there not be
    if not os.path.isdir(des):
        os.mkdir(des)
    # run algorithm if there is not result

    files = os.listdir(des)
    if not files:
        print('prediction...')
        from DICOM.AlgorithmAPI import main_interface
        main_interface.predictLungCancer(
            dcm_dir=base, model_dir='DICOM/models/', tgt_dir=des, tmp_dir='/tmp/plc/', pci_id='1')
        files = os.listdir(des)
        metadata = job.create_metadata()
        for file_name in files:
            if file_name.endswith('png') or file_name.endswith('jpg') or file_name.endswith('jpeg'):
                metadata.images.append(file_name)
            if file_name.endswith('csv'):
                with open('{des}/{file_name}'.format(des=des, file_name=file_name), 'r') as f:
                    metadata.features['results'] = []
                    reader = csv.DictReader(f)
                    for row in reader:
                        metadata.features['results'].append(row)
            if file_name.endswith('txt'):
                with open('{des}/{file_name}'.format(des=des, file_name=file_name), 'r') as f:
                    result = f.read()
                    metadata.fetures['total'] = result
        if not os.path.isfile('{des}/metadata.json'.format(des=des)):
            print('write metadata')
            with open('{des}/metadata.json'.format(des=des), 'w+') as f:
                j = serializer.dump(metadata)
                f.write(json.dumps(j))
        print('payload done')
        print('polling...')


def run_worker(host, port):
    conn = Redis(host=host, port=port)
    config = DICOMConfig()
    worker = GPUWorker(connection=conn, config=config, job_hash=HJ, sign_hash=HS,
                       queue_list=LQ, working_list=LW, done_list=LD, fail_list=LF)
    print('polling...')
    worker.polling(config=config, payload=payload)

if __name__ == '__main__':
    while True:
        try:
            run_worker(host='127.0.0.1', port='6389')
        except KeyboardInterrupt as e:
            break
        except Exception as e:
            print(e)
            with open('./worker.log', 'a+') as f:
                f.write(str(e))



import uuid
import os
import time
import requests




class BaseWorker(object):
    """docstring for BaseWorker"""
    conn = None
    config = None
    queue_list = None
    working_list = None
    done_list = None
    fail_list = None
    sign_hash = None
    job_hash = None
    id = None
    
    pid_file = 'worker.pid'

    headers = {'Authorization':'Basic YWRtaW46YWRtaW4xMjM0NTY='}

    def __init__(self, connection=None, config=None, job_hash=None, sign_hash=None, queue_list=None, working_list=None, done_list=None, fail_list=None):
        if connection:
            self.conn = connection
            self.job_hash = job_hash
            self.id = uuid.uuid1()
            self.queue_list = queue_list(
                connection=self.conn, job_hash=job_hash)
            self.working_list = working_list(
                connection=self.conn, job_hash=job_hash, list_name='work_list_{}'.format(self.id))
            self.sign_hash = sign_hash(self.conn)
            self.sign_hash.sign_in(self, self.working_list)
            self.done_list = done_list(connection=self.conn, job_hash=job_hash)
            self.fail_list = fail_list(connection=self.conn, job_hash=job_hash)
        if config:
            self.config = config
        if os.path.isfile(self.pid_file):
            # 宣告前任死亡,继承前任的工作
            with open(self.pid_file, 'r') as f:
                id = f.read()
                dead_working_list = working_list(
                    connection=self.conn, job_hash=self.job_hash, list_name='work_list_{}'.format(id))
                dead_working_list.id = id
                for job in dead_working_list:
                    print('find a dead job')
                    self.working_list.queue(job)
                self.sign_hash.sign_out(dead_working_list)

        with open(self.pid_file, 'w+') as f:
            f.write(str(self.id))

    def polling(self, payload=None, config=None):
        while True:
            try:
                if self.working_list.len() > 0 or self.queue_list.len() > 0:
                    self.management(payload, config)
                    job = self.working_list.pop_to(self.done_list)
                    data = {
                        'job_id':str(job.id),
                        'accession_id':str(job.content.accession_id),
                        'status':'finished',
                        'end_time':time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                    }
                    response = requests.put('{api}/{tasks}/{job_id}/?format={format}'
                            .format(tasks='tasks',
                                    job_id=job.id,
                                    format='json',
                                    api='http://127.0.0.1/api/v1'),headers=self.headers,json=data)
                    if response.status_code != 200:
                        raise Exception(response.url, response.content)
            except KeyboardInterrupt as e:
                os.remove(self.pid_file)
                self.sign_hash.sign_out(self)
                raise e
            else:
                continue

    def management(self, payload, config):
        raise NotImplementedError('management should implement')
